import { app, BrowserWindow, dialog, ipcMain, screen } from 'electron'
import * as isDev from 'electron-is-dev'
import { access, constants, PathLike, promises as fsPromises } from 'fs'
import * as path from 'path'
import * as usbDetect from 'usb-detection'
import { promisify } from 'util'

import { runUpdater } from './updater/Updater'
import { notifyUserPopup } from './userNotification'

// Using sudoPromptExec to clarify it's for elevated permissions
const sudoPromptExec = require('sudo-prompt').exec

const exec = promisify(require('child_process').exec)

const executeCommand = (command: string) => {
  return new Promise((resolve, reject) => {
    exec(command, (error: { message: any }, stdout: unknown, stderr: any) => {
      if (error) {
        reject({ error: error.message, stderr })
      } else if (stderr) {
        reject({ error: 'Error output received', stderr })
      } else {
        resolve(stdout)
      }
    })
  })
}

const SUPPORTED_DEVICES_STRING = [
  { vendorId: '20A0', productIds: ['42DB', '42DA'] }, // Storing hexadecimal values as strings
  { vendorId: '3603', productIds: ['0001', '0002', '0003', '0004'] }
]

const SUPPORTED_DEVICES = [
  { vendorId: 0x20a0, productIds: [0x42db, 0x42da] },
  { vendorId: 0x3603, productIds: [0x0001, 0x0002, 0x0003, 0x0004] }
]

// Handle creating/removing shortcuts on Windows when installing/uninstalling.
if (require('electron-squirrel-startup')) {
  app.quit()
}

const REACT_DEV_URL = 'http://localhost:3000'

// Global reference to mainWindow (necessary to prevent window from being garbage collected)
let mainWindow: BrowserWindow | undefined

const isElectron = 'electron' in process.versions
const isUsingAsar = isElectron && process.mainModule?.filename.includes('app.asar')
const fixPathForAsarUnpack = (path: string) =>
  isUsingAsar ? path.replace('app.asar', 'app.asar.unpacked') : path

const createRulesForLinux = async () => {
  const deviceArguments = SUPPORTED_DEVICES_STRING.reduce((acc, device) => {
    const ids = device.productIds.map(
      pid => `${encodeURIComponent(device.vendorId)} ${encodeURIComponent(pid)}`
    )
    return acc.concat(ids)
  }, []).join(' ')

  const checkCommand = `bash ${path.join(
    fixPathForAsarUnpack(__dirname),
    '../src/readRules.sh'
  )} ${deviceArguments}`

  try {
    const checkResult = await exec(checkCommand)
    console.log('Full output from check script:', checkResult.stdout)

    const results = checkResult.stdout
      .trim()
      .split('\n')
      .map((line: string) => line.trim())
    const changesRequired = results.includes('Changes required')

    if (changesRequired) {
      console.log('readRules is reporting that changes are required')

      // First notify the user with a popup that changes are required
      notifyUserPopup(
        'Update Required',
        'Changes to udev rules are required for your MuteMe device to function properly on Linux.',
        8000,
        true,
        'Learn More',
        'https://muteme.com/pages/linuxsetup'
      )

      await delay(7000) // Wait for the user to acknowledge the popup

      // Then ask if they want to perform the update
      const { response } = await dialog.showMessageBox({
        type: 'info',
        buttons: ['Update', 'Cancel'],
        defaultId: 0,
        title: 'Permission Required',
        message:
          'Would you like to update the udev rules now? This operation requires elevated permissions.',
        detail: 'Updating will ensure your device functions correctly.'
      })

      if (response === 0) {
        // If user clicks 'Update'
        const updateCommand = `bash ${path.join(
          fixPathForAsarUnpack(__dirname),
          '../src/updateRules.sh'
        )} ${deviceArguments}`
        console.log('Executing the update command with elevated permissions')
        const updateResult = await sudoPromptExec(updateCommand, { name: 'MuteMe' })
        console.log('Update script executed successfully:', updateResult)
      } else {
        console.log('User canceled the udev rule update.')
        notifyUserPopup(
          'Update Required To Use MuteMe',
          'Changes to udev rules are required for your MuteMe device to connect, you can choose to do this manually',
          12000,
          true,
          'Learn More',
          'https://muteme.com/pages/linuxsetup'
        )
      }
    } else {
      console.log('No changes to udev rules required.')
    }

    checkPulseAudio() // Check PulseAudio status
    return true
  } catch (error) {
    console.error('Failed to execute script:', error)
    return false
  }
}

interface ConnectedDevices {
  [key: string]: boolean
}

const createMainWindow = (): void => {
  let mainWindow = new BrowserWindow({
    width: 600,
    height: 750,
    resizable: false,
    frame: true,
    backgroundColor: '#282c34',
    titleBarStyle: 'hidden',
    webPreferences: {
      nodeIntegration: true,
      enableRemoteModule: true,
      preload: path.join(__dirname, 'preload.js'),
      backgroundThrottling: false
    }
  })

  const startURL = isDev
    ? REACT_DEV_URL
    : `file://${path.join(__dirname, '../static/build/index.html')}`
  mainWindow.loadURL(startURL)
  mainWindow.setMenuBarVisibility(false)

  usbDetect.startMonitoring()

  // Initialize an object to keep track of connected devices
  let connectedDevices: ConnectedDevices = {}

  SUPPORTED_DEVICES.forEach(device => {
    device.productIds.forEach(pid => {
      const eventString = `add:${device.vendorId}:${pid}`
      const removeString = `remove:${device.vendorId}:${pid}`

      usbDetect.on(eventString, device => {
        const deviceKey = `${device.vendorId}:${pid}`
        if (!connectedDevices[deviceKey]) {
          console.log(`[Electron]: plugged in MuteMe`, device)
          mainWindow?.webContents.send('usb-detected-muteme', device)
          connectedDevices[deviceKey] = true
        }
      })

      usbDetect.on(removeString, device => {
        const deviceKey = `${device.vendorId}:${pid}`
        if (connectedDevices[deviceKey]) {
          console.log(`[Electron]: unplugged MuteMe`, device)
          mainWindow?.webContents.send('usb-detected-unmuteme', device) // Notify frontend about the disconnection
          delete connectedDevices[deviceKey]
        }
      })
    })
  })

  if (process.platform === 'linux') {
    createRulesForLinux().then(async rulesExist => {
      console.log(`rules exist for linux: ${rulesExist}`)
      if (rulesExist) {
        SUPPORTED_DEVICES.forEach(async device => {
          const devices = await usbDetect.find(device.vendorId)
          devices
            .filter(d => device.productIds.includes(d.productId))
            .forEach(d => {
              mainWindow?.webContents.send('usb-detected-muteme', d)
            })
        })
      }
    })
  }

  mainWindow.on('closed', () => {
    usbDetect.stopMonitoring()
    connectedDevices = {}
    mainWindow = undefined
  })
}

const exitApp = () => {
  try {
    usbDetect.stopMonitoring()
  } catch (error) {
    console.log(error)
  }
  app.quit()
}

app.commandLine.appendSwitch('disable-renderer-backgrounding')

// Quit application when all windows are closed
app.on('window-all-closed', () => {
  // On macOS it is common for applications to stay open until the user explicitly quits
  // if (process.platform !== 'darwin') {
  exitApp()
  // }
})

app.on('activate', () => {
  // On macOS it is common to re-create a window even after all windows have been closed
  if (!mainWindow) {
    createMainWindow()
  }
})

const gotTheLock = app.requestSingleInstanceLock()

if (!gotTheLock) {
  exitApp()
} else {
  app.on('second-instance', (_event, _commandLine, _workingDirectory) => {
    // Someone tried to run a second instance, we should focus our window.
    if (mainWindow) {
      if (mainWindow.isMinimized()) mainWindow.restore()
      mainWindow.focus()
    }
  })

  // Create mainWindow, load the rest of the app, etc...
  app.whenReady().then(() => {
    createMainWindow()
  })
}

ipcMain.handle('log', (_event: any, ...args: any) => {
  console.debug(...args)
})

function delay(ms: number): Promise<void> {
  return new Promise(resolve => setTimeout(resolve, ms))
}

const checkPulseAudio = async () => {
  await delay(5000) // Delay for reassurance
  exec('pulseaudio --check', async (error: { code: number }, stdout: any, stderr: any) => {
    if (error) {
      if (error.code === 1) {
        console.log('PulseAudio is not running, attempting to start...')
        try {
          await exec('pulseaudio --start')
          console.log('PulseAudio started successfully.')
        } catch (startError) {
          console.error('Failed to start PulseAudio:', startError)
          notifyUserPopup(
            'PulseAudio Start Failed',
            'Failed to start PulseAudio. Please manually start it for full functionality.',
            10000,
            true,
            'Learn More',
            'https://muteme.com/pages/linuxsetup'
          )
        }
      } else {
        // Other errors (e.g., command not found)
        console.error('Failed to execute PulseAudio check:', error)
        notifyUserAboutPulseAudio()
      }
    } else {
      console.log('PulseAudio is running')
    }
  })
}

const notifyUserAboutPulseAudio = async () => {
  notifyUserPopup(
    'Pulseaudio Required',
    'Pulseaudio is required to control the microphone, without it you cannot mute or unmute the microphone',
    7000,
    true,
    'Learn More',
    'https://muteme.com/pages/linuxsetup'
  )
  await delay(7000) // Wait for the user to acknowledge the popup

  const { response } = await dialog.showMessageBox({
    type: 'info',
    buttons: ['Install', 'Cancel'],
    defaultId: 0,
    title: 'PulseAudio Missing',
    message:
      'PulseAudio is not installed or not running. Would you like us to install and enable it?',
    detail: 'Installing PulseAudio will allow for microphone controll',
    noLink: true
  })

  if (response === 0) {
    console.log('User opted to install PulseAudio.')
    installAndActivatePulseAudio()
  } else {
    console.log('User cancelled the PulseAudio installation.')
    notifyUserPopup(
      'Pulse Audio Required',
      'Pulse Audio is required to control microphone you can enable it manually',
      10000,
      true,
      'Learn More',
      'https://muteme.com/pages/linuxsetup'
    )
  }
}

const scriptPath = path.join(fixPathForAsarUnpack(__dirname), '../src/installPulseAudio.sh')

// Function to ensure the script is executable
const ensureScriptExecutable = async (scriptPath: PathLike) => {
  try {
    // Check if the script can be accessed
    await fsPromises.access(scriptPath, constants.F_OK)
    console.log('Script found, setting executable permissions...')

    // Command to change permissions
    const chmodCommand = `chmod 0755 ${scriptPath}`

    // Execute the chmod command with sudo
    const result = await sudoPromptExecPromise(chmodCommand, { name: 'MuteMe' })
    console.log('Script is now executable:', result.stdout)
  } catch (error) {
    console.error('Error setting script permissions:', error)
    throw new Error('Failed to set script permissions.')
  }
}

interface CommandResult {
  stdout: string
  stderr: string
}

// Wrapping sudoPromptExec to return a Promise with explicit types
const sudoPromptExecPromise = (command: string, options: any): Promise<CommandResult> => {
  return new Promise((resolve, reject) => {
    sudoPromptExec(command, options, (error: any, stdout: string, stderr: string) => {
      if (error) {
        reject({ error, stderr }) // Reject the Promise with error and stderr
      } else {
        resolve({ stdout, stderr }) // Resolve the Promise with stdout and stderr
      }
    })
  })
}

const installAndActivatePulseAudio = async () => {
  try {
    console.log('Initiating PulseAudio installation process...')

    // Ensure the script is executable
    await ensureScriptExecutable(scriptPath)

    // Proceed with the script execution
    const scriptExecutionCommand = `bash ${scriptPath}`
    console.log(`Executing script: ${scriptExecutionCommand}`)
    const scriptResult = await sudoPromptExecPromise(scriptExecutionCommand, { name: 'MuteMe' })
    console.log('Sudo Script execution result:', scriptResult.stdout)
    await executeCommand(`pulseaudio --start`)

    console.log('PulseAudio installation and configuration commands completed.')
    await delay(5000) // Delay for reassurance
    await checkPulseAudioPostInstallation()
  } catch (error) {
    console.error('An error occurred during the PulseAudio installation process:', error)
    notifyUserPopup(
      'Pulseaudio Installation Failed',
      'Failed to install or activate PulseAudio. Please try manually or contact support.',
      7000,
      true,
      'Learn More',
      'https://muteme.com/pages/linuxsetup'
    )
  }
}

const checkPulseAudioPostInstallation = async () => {
  exec('pulseaudio --check', (error: { code: number }, stdout: any, stderr: any) => {
    if (error) {
      console.error('Checking PulseAudio status failed:', error)
      console.log('Standard Output:', stdout) // Log stdout for more context
      console.error('Standard Error:', stderr) // Log stderr for potential errors

      const message =
        error.code === 1
          ? 'PulseAudio was installed but could not be started. Please manually start PulseAudio.'
          : 'Unable to verify if PulseAudio is running. Please check manually.'

      notifyUserPopup(
        'Pulseaudio Post-Installation Check Failed',
        `${message}`,
        7000,
        true,
        'Learn More',
        'https://muteme.com/pages/linuxsetup'
      )
    } else {
      console.log('PulseAudio is active and running after installation.')
      notifyUserPopup(
        'Pulseaudio Installed',
        'PulseAudio has been successfully installed and activated.',
        7500,
        true,
        'Learn More',
        'https://muteme.com/pages/linuxsetup'
      )
    }
  })
}
