// This is the entry point to the auto-updater, and should
// be required by the main electron process.

const { autoUpdater, ipcMain, dialog, app } = require('electron')

const winston = require('winston')

const logFormatter = winston.format.printf(info => {
  let { timestamp, level, code, stack, message, ...rest } = info

  // print out http error code w/ a space if we have one
  code = code ? ` ${code}` : ''
  // print the stack if we have it, message otherwise.
  message = stack || message

  return `> ${timestamp} ${level}${code}: ${message} ${
    Object.keys(rest).length !== 0 ? `\n${JSON.stringify(rest, null, 2)}` : ''
  }`
})

winston.configure({
  level: 'info',
  // put the errors formatter in the parent for some reason, only needed there:
  format: winston.format.errors({ stack: true }),
  transports: [
    new winston.transports.File({
      filename: `updater.log`,
      format: winston.format.combine(winston.format.timestamp(), logFormatter)
    })
  ]
})

const getPlatform = () => {
  switch (process.platform) {
    case 'win32':
      return process.arch === 'x64' ? 'win64' : 'win32'
    case 'darwin':
      return 'darwin'
  }
}

const runUpdater = mainWindow => {
  winston.log('info', `[Updater]: v${app.getVersion()} ================================`)

  autoUpdater.on('checking-for-update', info => {
    winston.log('info', `autoUpdater.on('checking-for-update')`, info)
    mainWindow.webContents.send('updaterEvent', {
      event: 'checking-for-update',
      data: info
    })
  })

  autoUpdater.on('update-available', info => {
    winston.log('info', `autoUpdater.on('update-available')`, info)
    mainWindow.webContents.send('updaterEvent', {
      event: 'update-available',
      data: info
    })
  })

  autoUpdater.on('update-not-available', info => {
    winston.log('info', `autoUpdater.on('update-not-available')`, info)
    mainWindow.webContents.send('updaterEvent', {
      event: 'update-not-available',
      data: info
    })
  })

  autoUpdater.on(
    'update-downloaded',
    (event, releaseNotes, releaseName, releaseDate, updateURL) => {
      // const info = { releaseNotes, releaseName, releaseDate, updateURL }
      mainWindow.webContents.send('updaterEvent', {
        event: 'update-downloaded',
        data: null
      })
      winston.log('info', `autoUpdater.on('update-downloaded')`)
    }
  )

  autoUpdater.on('error', info => {
    winston.log('info', `autoUpdater.on('error') fucker: ${info}`)
    winston.log('error', info)
    mainWindow.webContents.send('updaterEvent', {
      event: 'error',
      data: info
    })
  })

  ipcMain.handle('autoUpdater-showDialog', (_event, _info) => {
    // const { releaseNotes, releaseName } = info
    const dialogOpts = {
      type: 'info',
      buttons: ['Restart', 'Later'],
      title: 'Application Update',
      // message: process.platform === 'win32' ? releaseNotes : releaseName,
      detail: 'A new version has been downloaded. Restart the application to apply the updates.'
    }

    dialog.showMessageBox(dialogOpts).then(returnValue => {
      if (returnValue.response === 0) autoUpdater.quitAndInstall()
    })
  })

  try {
    const platform = getPlatform()
    const server = 'http://muteme.io'
    const url = `${server}/update/${platform}/${app.getVersion()}`
    autoUpdater.setFeedURL(url)
    winston.log('info', `initiating version check at ${url}`)
    autoUpdater.checkForUpdates()
  } catch (error) {
    winston.log('info', `checkForUpdates Error ${error}`)
    winston.log('error', error)
    mainWindow.webContents.send('updaterEvent', {
      event: 'checking-for-update',
      data: error
    })
  }
}

exports.runUpdater = runUpdater
